#include <math.h>
#include <pc.h>
#include "mode15k.h"


extern CRTCPARAM modeF5;
extern CRTCPARAM modeF6;
extern CRTCPARAM modeF7;
extern CRTCPARAM modeF8;
extern CRTCPARAM modeF9;
extern CRTCPARAM modeF10;
extern CRTCPARAM modeF11;
extern CRTCPARAM modeF12;


static void vesa_scroll(int x, int y);

int center_x = 0;
int center_y = 0;
int interlaceratio = 0;
int compositesync  = 0;


#define GET_320x240_60_15KHz   1
#define GET_512x384_60_24KHz   2
#define GET_640x480_60_31KHz   3
#define GET_512x384_60_31KHz   4
#define GET_F5	5
#define GET_F6	6
#define GET_F7	7
#define GET_F8	8
#define GET_F9	9
#define GET_F10 10
#define GET_F11 11
#define GET_F12 12

#define MODE15KHz	(1 << 0)
#define MODE24KHz	(1 << 1)
#define MODE31KHz	(1 << 2)
#define MODE35KHz	(1 << 3)
#define OVER31KHz	(1 << 16)

#define _H		1
#define _V		2

static void SetMiscParam(CRTCPARAM *cp);
static void AdjustCenter(CRTCPARAM *cp);



static CRTCPARAM mode320x240_60_15KHz = {
// HTotal Disp BSta SSta SEnd BEnd VTotal Disp BSta SSta SEnd BEnd
	  432, 320, 320, 360, 400, 424,   263, 240, 240, 241, 243, 262,

// width height HPlr VPlr DblScn Intrlc Iratio
	 320,	240,   1,	1,	   0,	  0,   100,

// dotclkHz  horzHz  vertHz(in 0.01)
	6816960,  15780,   6000
};



static CRTCPARAM mode512x384_60_24KHz = {
// HTotal Disp BSta SSta SEnd BEnd VTotal Disp BSta SSta SEnd BEnd
      712, 512, 512, 600, 640, 704,   414, 384, 384, 383, 387, 413,

// width height HPlr VPlr DblScn Intrlc Iratio 
     512,   384,   1,   1,     0,     0,     0, 

// dotclkHz  horzHz  vertHz(in 0.01)
   17686080,  24840,   6000
};



static CRTCPARAM mode640x480_60_31KHz = {
// HTotal Disp BSta SSta SEnd BEnd VTotal Disp BSta SSta SEnd BEnd
      800, 640, 640, 672, 768, 792,   525, 480, 480, 481, 484, 524,

// width height HPlr VPlr DblScn Intrlc Iratio 
     640,   480,   1,   1,     0,     0,     0, 

// dotclkHz  horzHz  vertHz(in 0.01)
   25200000,  31500,   6000
};


static CRTCPARAM mode512x384_60_31KHz = {
// HTotal Disp BSta SSta SEnd BEnd VTotal Disp BSta SSta SEnd BEnd
	  648, 512, 512, 544, 584, 640,   525, 384, 384, 435, 438, 524,

// width height HPlr VPlr DblScn Intrlc Iratio
	 512,	384,   1,	0,	   0,	  0,   100,

// dotclkHz  horzHz  vertHz(in 0.01)
   20412000,  31500,   6000
};






static CRTCPARAM modedummy = {
// HTotal Disp BSta SSta SEnd BEnd VTotal Disp BSta SSta SEnd BEnd
		0,	 0,   0,   0,	0,	 0, 	0,	 0,   0,   0,	0,	 0,

// width height HPlr VPlr DblScn Intrlc Iratio
	   0,	  0,   0,	0,	   0,	  0,	 0,

// dotclkHz  horzHz  vertHz(in 0.01)
		  0,	  0,	  0
};



void logerror(char* txt, ...)
{


}



static CRTCPARAM g_cp;

void SetCRTCPARAM(CRTCPARAM* cp)
{
	g_cp = modedummy;
	g_cp.HTotal  = cp->HTotal;
	g_cp.HDisp	 = cp->HDisp;
	g_cp.HSStart = cp->HSStart;
	g_cp.HSEnd	 = cp->HSEnd;
	g_cp.VTotal  = cp->VTotal;
	g_cp.VDisp	 = cp->VDisp;
	g_cp.VSStart = cp->VSStart;
	g_cp.VSEnd	 = cp->VSEnd;

	g_cp.dotclockHz = cp->dotclockHz;
	g_cp.horzHz 	= cp->horzHz;
	g_cp.vert001Hz	= cp->vert001Hz;


	g_cp.width	= cp->HDisp;
	g_cp.height = cp->VDisp;
	g_cp.hpolarity	= cp->hpolarity;
	g_cp.vpolarity	= cp->vpolarity;
	g_cp.doublescan = cp->doublescan;
	g_cp.interlace	= cp->interlace;

}



void GetCRTCPARAM(CRTCPARAM* cp, int num)
{
	switch(num){
		case GET_320x240_60_15KHz:
			*cp = mode320x240_60_15KHz; break;
		case GET_512x384_60_24KHz:
			*cp = mode512x384_60_24KHz; break;
		case GET_640x480_60_31KHz:
			*cp = mode640x480_60_31KHz; break;
		case GET_512x384_60_31KHz:
			*cp = mode512x384_60_31KHz; break;
		case GET_F5:
			*cp = modeF5;  break;
		case GET_F6:
			*cp = modeF6;  break;
		case GET_F7:
			*cp = modeF7;  break;
		case GET_F8:
			*cp = modeF8;  break;
		case GET_F9:
			*cp = modeF9;  break;
		case GET_F10:
			*cp = modeF10; break;
		case GET_F11:
			*cp = modeF11; break;
		case GET_F12:
			*cp = modeF12; break;

		default:
			*cp = mode640x480_60_31KHz;
	}

	cp->HTotal	&= ~0x7;
	cp->HDisp	&= ~0x7;
	cp->HSStart &= ~0x7;
	cp->HSEnd	&= ~0x7;
}




int GetMode15KHz(CRTCPARAM *cp)
{
	*cp = g_cp;
	AdjustCenter(cp);
	SetMiscParam(cp);

	return 1;
}





static void SetMiscParam(CRTCPARAM *cp)
{
	if(interlaceratio == 0)  interlaceratio = 100;
	if(interlaceratio < 0)	 interlaceratio = 0;
	if(interlaceratio > 200) interlaceratio = 200;
	cp->interlaceratio = interlaceratio;
}

static void AdjustCenter(CRTCPARAM *cp)
{
	int center_x_temp;
	int center_y_temp;
	int d0;

	center_x_temp = center_x & ~0x7;
	if(center_x < 0){
		if((cp->HSStart + center_x) < (cp->HDisp + 8)){
			d0 = cp->HSStart;
			cp->HSStart = cp->HDisp + 8;
			cp->HSEnd  -= (d0 - cp->HSStart);
		}else{
			cp->HSStart = cp->HSStart + center_x;
			cp->HSEnd	= cp->HSEnd   + center_x;
		}
	}else{
		if((cp->HSEnd + center_x) > (cp->HTotal - 16)){
			d0 = cp->HSEnd;
			cp->HSEnd	 = cp->HTotal - 16;
			cp->HSStart += (cp->HSEnd - d0);
		}else{
			cp->HSEnd	= cp->HSEnd   + center_x;
			cp->HSStart = cp->HSStart + center_x;
		}
	}

	center_y_temp = center_y;
	if(center_y < 0){
		if((cp->VSStart + center_y) < (cp->VDisp + 1)){
			d0 = cp->VSStart;
			cp->VSStart = cp->VDisp + 1;
			cp->VSEnd  -= (d0 - cp->VSStart);
		}else{
			cp->VSStart = cp->VSStart + center_y;
			cp->VSEnd	= cp->VSEnd   + center_y;
		}
	}else{
		if((cp->VSEnd + center_y) > (cp->VTotal - 2)){
			d0 = cp->VSEnd;
			cp->VSEnd	 = cp->VTotal - 2;
			cp->VSStart += (cp->VSEnd - d0);
		}else{
			cp->VSEnd	= cp->VSEnd   + center_y;
			cp->VSStart = cp->VSStart + center_y;
		}
	}
}




#if 0
void GetCrtcParam(CRTCPARAM *cp, MiscParam *mc, _Register *rg, int registerlen)
{
	BYTE port[0x20];
	int d0;
	int d1;

	d0 = 0;
	while(d0 < registerlen){
		port[ rg[d0].index ] = rg[d0].value;
		d0 += 1;
	}


/*										*/
/*				 CRTC					*/
/*										*/

	cp->HTotal	= (port[0x00] + 5) * 8;
	cp->HDisp	= (port[0x01] + 1) * 8;

	d0 = port[0x02];
	cp->HBStart = d0 * 8;
	d1 = (port[0x03] & 0x1F) | BitMov(0, 5, port[0x05], 7);
	while( (d0 & 0x3F) != (d1 & 0x3F) ){
		d0 += 1;
	}
	cp->HBEnd	= d0 * 8;

	d0 = port[0x04];
//	  cp->HSStart = d0 * 8;
	cp->HSStart = d0 * 8 + mc->hskew;
	d1 = (port[0x05] & 0x1F);
	while( (d0 & 0x1F) != (d1 & 0x1F) ){
		d0 += 1;
	}
	cp->HSEnd	= d0 * 8;

	cp->VTotal	= (port[0x06] | BitMov(0, 8, port[0x07], 0) |
								BitMov(0, 9, port[0x07], 5) ) + 2;
	cp->VDisp	= (port[0x12] | BitMov(0, 8, port[0x07], 1) |
								BitMov(0, 9, port[0x07], 6) ) + 1;
	cp->VBStart = (port[0x15] | BitMov(0, 8, port[0x07], 3) |
								BitMov(0, 9, port[0x09], 5) );
	d0 = cp->VBStart;
	d1 = port[0x16];
	while( (d0 & 0xFF) != (d1 & 0xFF) ){
		d0 += 1;
	}

//.    cp->VBEnd = d0;
//.    cp->VBEnd = cp->VTotal - 16;


	cp->VSStart = (port[0x10] | BitMov(0, 8, port[0x07], 2) |
								BitMov(0, 9, port[0x07], 7) );
	d0 = cp->VSStart;
	d1 = port[0x11] & 0x0F;
	while( (d0 & 0x0F) != (d1 & 0x0F) ){
		d0 += 1;
	}
	cp->VSEnd = d0;
//.
	cp->VBEnd = cp->VSEnd + 1;
//.



/*										*/
/*				 MISC					*/
/*										*/

	cp->width  = mc->x;
	cp->height = mc->y;
	cp->depth  = 0;

	cp->dotclockHz = (long)(mc->dotclock * 1000 * 1000);
	cp->horzHz	   = (long)(mc->horzKHz * 1000);
	cp->vert001Hz  = (long)((mc->vertHz + 0.005) * 100);

	cp->hpolarity  = mc->hpolarity;
	cp->vpolarity  = mc->vpolarity;
	cp->doublescan = (port[0x09] >> 7) & 0x01;
	cp->interlace  = mc->interlace;

	if(interlaceratio < 0)	 interlaceratio = 0;
	if(interlaceratio > 200) interlaceratio = 200;
	cp->interlaceratio = interlaceratio;
	cp->compositesync  = compositesync;

	cp->sis_numerator	= mc->numerator;
	cp->sis_denumerator = mc->denumerator;
	cp->sis_divider 	= mc->divider;
	cp->sis_postscalar	= mc->postscalar;
	cp->sis_seqdivider	= mc->seqdivider;
}
#endif




void SetGenericVGARegister(CRTCPARAM *cp)
{
	int d0;
	int d1;
	int HDispSkew  = 0;
	int HSyncDelay = 0;

	// Unlock
	SetCrtcReg(0x11, BitMov(GetCrtcReg(0x11), 7, 0, 0));

	// HTotal
	d0 = cp->HTotal;
	d0 = (d0 / 8) - 5;
	SetCrtcReg(0x00, d0 & 0xFF);

	// HDisp
	d0 = cp->HDisp;
	d0 = (d0 / 8) - 1;
	SetCrtcReg(0x01, d0 & 0xFF);

	// HBlankStart
	d0 = cp->HBStart;
	d0 = d0 / 8;
	SetCrtcReg(0x02, d0 & 0xFF);

	// HBlankEnd
	d0 = cp->HBEnd;
	d0 = d0 / 8;
	d1 = GetCrtcReg(0x03);
	d1 = BitMov(d1, 0, d0, 0);
	d1 = BitMov(d1, 1, d0, 1);
	d1 = BitMov(d1, 2, d0, 2);
	d1 = BitMov(d1, 3, d0, 3);
	d1 = BitMov(d1, 4, d0, 4);
	d1 = BitMov(d1, 5, HDispSkew, 0);	// horizontal disp skew
	d1 = BitMov(d1, 6, HDispSkew, 1);	// horizontal disp skew
	d1 = BitMov(d1, 7, 1, 0);			// compatible read
	SetCrtcReg(0x03, d1);
	SetCrtcReg(0x05, BitMov(GetCrtcReg(0x05), 7, d0, 5));

	// HSyncStart
	d0 = cp->HSStart;
	d0 = d0 / 8;
	SetCrtcReg(0x04, d0 & 0xFF);

	// HSyncEnd
	d0 = cp->HSEnd;
	d0 = d0 / 8;
	d1 = GetCrtcReg(0x05);
	d1 = BitMov(d1, 0, d0, 0);
	d1 = BitMov(d1, 1, d0, 1);
	d1 = BitMov(d1, 2, d0, 2);
	d1 = BitMov(d1, 3, d0, 3);
	d1 = BitMov(d1, 4, d0, 4);
	d1 = BitMov(d1, 5, HSyncDelay, 0);
	d1 = BitMov(d1, 6, HSyncDelay, 1);
	SetCrtcReg(0x05, d1);

	// VTotal
	d0 = cp->VTotal;
	d0 = d0 - 2;
	SetCrtcReg(0x06, d0 & 0xFF);
	SetCrtcReg(0x07, BitMov(GetCrtcReg(0x07), 0, d0, 8));
	SetCrtcReg(0x07, BitMov(GetCrtcReg(0x07), 5, d0, 9));

	// VDisp
	d0 = cp->VDisp;
	d0 = d0 - 1;
	SetCrtcReg(0x12, d0 & 0xFF);
	SetCrtcReg(0x07, BitMov(GetCrtcReg(0x07), 1, d0, 8));
	SetCrtcReg(0x07, BitMov(GetCrtcReg(0x07), 6, d0, 9));


	// VBlankStart
	d0 = cp->VBStart;
	SetCrtcReg(0x15, d0 & 0xFF);
	SetCrtcReg(0x07, BitMov(GetCrtcReg(0x07), 3, d0, 8));
	SetCrtcReg(0x09, BitMov(GetCrtcReg(0x09), 5, d0, 9));

	// VBlankEnd
	d0 = cp->VBEnd;
	SetCrtcReg(0x16, d0 & 0xFF);

	// VSyncStart
	d0 = cp->VSStart;
	SetCrtcReg(0x10, d0 & 0xFF);
	d1 = GetCrtcReg(0x07);
	d1 = BitMov(d1, 2, d0,	8);  // bit 8
	d1 = BitMov(d1, 7, d0,	9);  // bit 9
	SetCrtcReg(0x07, d1);

	// VSyncEnd
	d0 = cp->VSEnd;
	d0 = d0 & 0x0F;
	d1 = GetCrtcReg(0x11) & 0xF0;
	d1 = d1 | d0;
	SetCrtcReg(0x11, d1);

	// Double Scan Dis.
	SetCrtcReg(0x09, BitMov(GetCrtcReg(0x09), 7, 0, 0));

}



long CalcDotclock(long Vert001Hz, int HTotal, int VTotal, int interlace)
{
	double f0;

	f0 = (double)VTotal;
	if(interlace != 0){
		f0 += 0.5;
	}
	f0 = f0 * HTotal;
	f0 = f0 * Vert001Hz;
	f0 = f0 / 100.0;

	return (long)f0;
}

long CalcHorizontalHz(long dotclockHz, int HTotal)
{
	return dotclockHz / HTotal;
}

long CalcVert001Hz(long dotclockHz, int HTotal, int VTotal, int interlace)
{
	double f0;

	f0 = (double)VTotal;
	if(interlace != 0){
		f0 += 0.5;
	}
	f0 = f0 * HTotal;
	f0 = dotclockHz / f0;
	f0 = f0 * 100.0;

	return (long)f0;
}



void DispCrtcParam(CRTCPARAM *cp)
{
	logerror("DotClock..... %.4f MHz\n",
						(float)(cp->dotclockHz / 1000) / 1000.0);
	logerror("HTotal....... %d\n", cp->HTotal);
	logerror("HDispEnd..... %d\n", cp->HDisp);
	logerror("HBlankStart.. %d\n", cp->HBStart);
	logerror("HSyncStart... %d\n", cp->HSStart);
	logerror("HSyncEnd..... %d\n", cp->HSEnd);
	logerror("HBlankEnd.... %d\n", cp->HBEnd);
	logerror("VTotal....... %d\n", cp->VTotal);
	logerror("VDispEnd..... %d\n", cp->VDisp);
	logerror("VBlankStart.. %d\n", cp->VBStart);
	logerror("VSyncStart... %d\n", cp->VSStart);
	logerror("VSyncEnd..... %d\n", cp->VSEnd);
	logerror("VBlankEnd.... %d\n", cp->VBEnd);
	logerror("Horz......... %.2f KHz\n",
						(float)cp->horzHz / 1000.0);
	logerror("Vert......... %.2f Hz\n",
						(float)cp->vert001Hz / 100.0);

	if(cp->hpolarity == 0){
		logerror("HorzSync..... Positive (Low  -> High)\n");
	}else{
		logerror("HorzSync..... Negative (High -> Low)\n");
	}
	if(cp->vpolarity == 0){
		logerror("VertSync..... Positive (Low  -> High)\n");
	}else{
		logerror("VertSync..... Negative (High -> Low)\n");
	}

	if(cp->interlace == 0){
		logerror("non interlaced mode\n");
	}else{
		logerror("interlaced mode\n");
	}

}


void SetHVPolarity(int HPolar, int VPolar)
{
	BYTE d0;

	if(HPolar != 0) HPolar = 1;
	if(VPolar != 0) VPolar = 1;

	d0 = _in(0x3CC);
	d0 = d0 & 0x3F;
	d0 = d0 | (VPolar << 7) | (HPolar << 6);
	_out(0x3C2, d0);
}





DWORD BitMov(DWORD dstdata, int dstbitpos, DWORD srcdata, int srcbitpos)
{
	DWORD d0;
	DWORD d1;

	d1 = srcdata >> srcbitpos;
	d1 = d1 & 1;
	d0 = dstdata & ~(1 << dstbitpos);
	d1 = d1 << dstbitpos;
	d0 = d0 | d1;
	return d0;
}



void SetSeqReg(BYTE __idx, BYTE data)
{
	_out(0x3c4, __idx);
	_out(0x3c5, data);
}

BYTE GetSeqReg(BYTE __idx)
{
	BYTE d0;

	_out(0x3c4, __idx);
	d0 = _in(0x3c5);
	return d0;
}


BYTE GetAtrbData(BYTE __idx)
{
	_in(0x3da);
	_out(0x3c0, __idx | (1 << 5));
	return _in(0x3c1);
}

void SetAtrbData(BYTE __idx, BYTE data)
{
	_in(0x3da);
	_out(0x3c0, __idx | 0x20);
	_out(0x3c0, data);
}

void SetCrtcReg(BYTE __idx, BYTE data)
{
	_out(0x3d4, __idx);
	_out(0x3d5, data);
}

BYTE GetCrtcReg(BYTE __idx)
{
	BYTE d0;

	_out(0x3d4, __idx);
	d0 = _in(0x3d5);
	return d0;
}



int GetSeqDivider(void)
{
	int d0;

	d0	= (GetSeqReg(0x01) & 0x08) >>3;
	d0	= d0 + 1;
	return d0;
}

void SetSeqDivider(int d0)
{
	int d1;

	d0 -= 1;
	if(d0 != 0) d0 = 1 << 3;
	d1 = GetSeqReg(0x01) & ~0x08;
	d1 = d1 | d0;
	SetSeqReg(0x01, d1);
}


void _out(int edx, BYTE al)
{
	outportb(edx, al);
}

BYTE _in(int edx)
{
	return inportb(edx);
}


void _outb(int edx, BYTE al)
{
	outportb(edx, al);
}
BYTE _inb(int edx)
{
	return inportb(edx);
}

void _outw(int edx, WORD ax)
{
	outportw(edx, ax);
}
WORD _inw(int edx)
{
	return inportw(edx);
}

void _outl(int edx, DWORD eax)
{
	outportl(edx, eax);
}
DWORD _inl(int edx)
{
	return inportl(edx);
}



void DisableGenerateSignal(void)
{
	BYTE d0;

	d0 = GetCrtcReg(0x17);
	d0 = d0 & 0x7F;
	SetCrtcReg(0x17, d0);
}

void EnableGenerateSignal(void)
{
	BYTE d0;

	d0 = GetCrtcReg(0x17);
	d0 = d0 | 0x80;
	SetCrtcReg(0x17, d0);
}
