#include "mamalleg.h"
#include "driver.h"
#include <dos.h>
#include <conio.h>
#include <stdlib.h>
#include "ticker.h"
#include "wss.h"

//.s
void CLIB_DECL logerror(const char *text,...);
//.e




static int num_used_opl;
int nominal_sample_rate;
int soundcard, usestereo;
int attenuation = 0;
static int master_volume = 256;

static int stream_playing;
static INT16 *stream_cache_data;
static int stream_cache_len;
static int stream_cache_stereo;




extern float fps_value;

int actual_sample_rate = 0;
int samples_per_vsync = 0;
int _8bit_shift_count = 8;
int reverse_stereo = 0;
int sb_cursor_offset = 0;
int oplbaseindex = 0;
int oplbaseio = 0x388;


float frames_per_second(void)
{
	float rate;

	if(fps_value == 0.0){
		rate = Machine->drv->frames_per_second;
	}else{
		rate = fps_value;
	}
	return rate;
}


void set_opl_base(void)
{
	switch(oplbaseindex){
		case 0:
			oplbaseio = 0x388;
			break;
		case 1:
			oplbaseio = 0x398;
			break;
		case 2:
			oplbaseio = 0x220;
			break;
		case 3:
			oplbaseio = 0x230;
			break;
		case 4:
			oplbaseio = 0x240;
			break;
		case 5:
			oplbaseio = 0x250;
			break;
		case 6:
			oplbaseio = 0x260;
			break;

		default:
			oplbaseio = 0x388;
	}
	logerror("OPLbase 0x%x\n", oplbaseio);
}



int msdos_init_seal (void)
{
	return 0;
}




int msdos_init_sound(void)
{
	int i;
	int rate;

	w_set_sb_cursor_offset(sb_cursor_offset);

	if (soundcard == -1)
	{

		printf("\nSelect the audio device:\n");
		printf("  0. Eternal Silence\n");
		printf("  1. Sound Blaster\n");
		printf("  4. Ultrasound Max\n");
		printf("  5. Ultrasound\n");
		printf("  6. Windows Sound System\n");
		printf("  7. Ensoniq Soundscape\n");
		printf("  8. ESS Audiodrive\n");
		printf("  9. Sound Blaster Pro (interrupt only)\n");
		printf("\n");

		i = getch();
		soundcard = i - '0';
		if ( !((0 <= soundcard) && (soundcard <= 9)) )
		{
			printf("audio initialization failed\n");
			soundcard = -1;
			return 1;
		}

	}

	if(soundcard != 0)
		set_opl_base();

	if(soundcard == 0)
		rate = (int)((frames_per_second() * 10) + 0.5);
	else
		rate = Machine->sample_rate;

	if( w_sound_device_init(soundcard, rate) == _FALSE ){
		printf("%s\n", w_get_error_message() );
		printf("audio initialization failed\n");
		soundcard = -1;
		return 1;
	}
	w_set_8bitpcm_shift_count(_8bit_shift_count);
	if(reverse_stereo) w_reverse_stereo(_TRUE);

//	  printf("%s%7d Hz\n", w_get_device_name(), (int)w_get_nominal_sample_rate());
	logerror("%s%7d Hz\n", w_get_device_name(), (int)w_get_nominal_sample_rate());

	/* update the Machine structure to reflect the actual sample rate */
	osd_set_mastervolume(attenuation);	/* set the startup volume */

	w_set_watermark(2.2, w_get_nominal_sample_rate()/60);

	if(actual_sample_rate == 0){
		actual_sample_rate	 = w_get_actual_sample_rate();
		Machine->sample_rate = actual_sample_rate;
		logerror("actual sample rate: %d\n", actual_sample_rate);
		if(actual_sample_rate == 0){
			printf("device not found\n");
			printf("audio initialization failed\n");
			soundcard = -1;
			return 1;
		}
	}else{
		Machine->sample_rate = actual_sample_rate;
		logerror("actual sample rate: %d\n", actual_sample_rate);
	}

	w_set_watermark( 2.2, actual_sample_rate/frames_per_second() );

	logerror("Machine->sample_rate: %d\n", Machine->sample_rate);

	num_used_opl = 0;
	return 0;
}






void msdos_shutdown_sound(void)
{
	int chip,n;

	if (Machine->sample_rate != 0)
	{
		w_sound_device_exit();
	}

	for(chip=0;chip<num_used_opl;chip++)
	{
		/* silence the OPL */
		for (n = 0x40;n <= 0x55;n++)
		{
			osd_opl_control(chip,n);
			osd_opl_write(chip,0x3f);
		}
		for (n = 0x60;n <= 0x95;n++)
		{
			osd_opl_control(chip,n);
			osd_opl_write(chip,0xff);
		}
		for (n = 0xa0;n <= 0xb0;n++)
		{
			osd_opl_control(chip,n);
			osd_opl_write(chip,0);
		}
	}
}







/* global sample tracking */
static UINT32 samples_per_frame;
static UINT32 samples_left_over;
static UINT32 samples_this_frame;


int osd_start_audio_stream(int stereo)
{
	extern int video_sync;

	if (Machine->sample_rate == 0) return 0;

	if (stereo) stereo = 1; /* make sure it's either 0 or 1 */

	stream_cache_stereo = stereo;

	if(video_sync && (samples_per_vsync != 0)){
		samples_this_frame = samples_per_vsync;
		logerror("vsync is on\n");
		logerror("force to generate %d samples per frame\n", samples_per_vsync);
	}else{
		/* determine the number of samples per frame */
		samples_per_frame = (Machine->sample_rate << 12) / frames_per_second();

		/* compute how many samples to generate this frame */
		samples_left_over  = samples_per_frame;
		samples_this_frame = (UINT32)(samples_left_over >> 12);
		samples_left_over  = samples_left_over & 0xFFF;
	}

	stream_playing = 1;

	return samples_this_frame;
}




void osd_stop_audio_stream(void)
{
	w_clear_buffer();
}



static void updateaudiostream(void)
{
	extern int throttle;
	extern int video_sync;
	INT16 *data = stream_cache_data;
	int stereo = stream_cache_stereo;
	int len = stream_cache_len;
	int d0;

	if (throttle && !video_sync)   /* sync with audio only when speed throttling is not turned off */
	{
		profiler_mark(PROFILER_IDLE);
		while(1){
			d0 = w_get_watermark_status();
			if( (d0 == -1) || (d0 == 0) ) break;
			if(d0 == -2){
//				  w_reset_watermark();
				w_reset_watermark_ex(len);
				break;
			}
		}
		profiler_mark(PROFILER_END);
	}

	if(video_sync && (samples_per_vsync != 0)){
		w_lock_mixing_buffer(0);
	}else{
		w_lock_mixing_buffer(len);
	}
	if(stereo){
		w_mixing_stereo(data, len, 256, 256);
	}else{
		w_mixing(data, len, 256, 256);
	}
	w_unlock_mixing_buffer();
}








int osd_update_audio_stream(INT16 *buffer)
{
	extern int video_sync;

	stream_cache_data = buffer;
	stream_cache_len  = samples_this_frame;

	if(!video_sync){
		/* compute how many samples to generate next frame */
		samples_left_over += samples_per_frame;
		samples_this_frame = (UINT32)(samples_left_over >> 12);
		samples_left_over  = samples_left_over & 0xFFF;
	}else{
		if(samples_per_vsync != 0){
			samples_this_frame = samples_per_vsync;
		}else{
			samples_this_frame = w_get_next_req();
		}
	}

	return samples_this_frame;
}

int msdos_update_audio(void)
{
	if (Machine->sample_rate == 0 || stream_cache_data == 0) return 0;

	profiler_mark(PROFILER_MIXER);
	updateaudiostream();
	profiler_mark(PROFILER_END);
	return 1;
}

/* attenuation in dB */
void osd_set_mastervolume(int _attenuation)
{
	float volume;

	if (_attenuation > 0) _attenuation = 0;
	if (_attenuation < -32) _attenuation = -32;

	attenuation = _attenuation;

	volume = 256.0; /* range is 0-256 */
	while (_attenuation++ < 0)
		volume /= 1.122018454;	/* = (10 ^ (1/20)) = 1dB */


	master_volume = volume;

	w_set_master_volume(master_volume);
}


int osd_get_mastervolume(void)
{
	return attenuation;
}

void osd_sound_enable(int enable_it)
{
	if (enable_it)
	{
		w_set_master_volume(master_volume);
	}
	else
	{
		w_set_master_volume(0);
		w_clear_buffer();
	}
}



/* linux sound driver opl3.c does a so called tenmicrosec() delay */
static void tenmicrosec(void)
{
	int i;
	for (i = 0; i < 16; i++)
		inportb(0x80);
}

//#define MAX_OPLCHIP 2  /* SOUND BLASTER 16 or compatible ?? */
#define MAX_OPLCHIP 1  /* SOUND BLASTER pro compatible ??  */

void osd_opl_control(int chip,int reg)
{
	if (Machine->sample_rate == 0) return;
	if (soundcard == 0) return;

	if (chip >= MAX_OPLCHIP ) return;
	tenmicrosec();
	outportb(oplbaseio + (chip*2), reg);
}

void osd_opl_write(int chip,int data)
{
	if (Machine->sample_rate == 0) return;
	if (soundcard == 0) return;

	if (chip >=MAX_OPLCHIP ) return;
	tenmicrosec();
	outportb( (oplbaseio+1) + (chip*2), data);

	if(chip >= num_used_opl) num_used_opl = chip+1;
}
